/*
 * Copyright (c) 2023 Marceline Cramer
 * Copyright (c) 2023 Emma Tebibyte <emma@tebibyte.media>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

use std::{
    env::args,
    fs::File,
    io::{
        Read,
        stdout,
        Stdout,
        Write,
    },
    os::fd::FromRawFd,
};

use crossterm::{
    cursor,
    event::{ Event, KeyCode, KeyEvent, read },
    terminal,
    ExecutableCommand,
    Result
};
use ropey::Rope;
use yacexits::{ exit, EX_DATAERR, EX_UNAVAILABLE };

struct Buffer {
    pub text: Rope,
}

impl Buffer {
    pub fn from_str(text: &str) -> Self {
        Self {
            text: Rope::from_str(text),
        }
    }

    pub fn draw(&self, scroll: Cursor, out: &mut (impl ExecutableCommand + Write)) -> Result<u32> {
        let (cols, rows) = terminal::size()?;
        let lr_width = self.text.len_lines().ilog10() + 1;
        let gutter_width = lr_width + 1;
        let text_width = cols as usize - gutter_width as usize;

        out.execute(cursor::MoveTo(0, 0))?;

        for (row, line) in (0..rows).zip(self.text.lines_at(scroll.line)) {
            // only the last line is empty and should be skipped
            if line.len_chars() == 0 {
                break;
            }

            let row = row as usize + scroll.line;
            write!(out, "{:width$} ", row, width = lr_width as usize)?;

            let lhs = scroll.column;
            let width = line.len_chars().saturating_sub(1); // lop off whitespace
            if lhs < width {
                let window = text_width.min(width - lhs);
                let rhs = lhs + window;
                write!(out, "{}", line.slice(lhs..rhs))?;
            }

            out.execute(cursor::MoveToNextLine(1))?;
        }

        Ok(gutter_width)
    }

    pub fn clamped_cursor(&self, cursor: Cursor) -> Cursor {
        Cursor {
            line: cursor.line,
            column: cursor
                .column
                .min(self.text.line(cursor.line).len_chars() - 1),
        }
    }

    pub fn cursor_to_char(&self, cursor: Cursor) -> usize {
        let cursor = self.clamped_cursor(cursor);
        self.text.line_to_char(cursor.line) + cursor.column
    }

    pub fn move_cursor(&self, cursor: &mut Cursor, direction: Direction, enable_linewrap: bool) {
        *cursor = self.clamped_cursor(*cursor);
        match direction {
            Direction::Left => {
                if cursor.column > 0 {
                    cursor.column -= 1;
                } else if enable_linewrap && cursor.line > 0 {
                    cursor.line -= 1;
                    let line = self.text.line(cursor.line);
                    cursor.column = line.len_chars() - 1;
                }
            }
            Direction::Down => {
                if cursor.line + 2 < self.text.len_lines() {
                    cursor.line += 1;
                }
            }
            Direction::Up => {
                if cursor.line > 0 {
                    cursor.line -= 1;
                }
            }
            Direction::Right => {
                let line = self.text.line(cursor.line);
                if cursor.column + 2 > line.len_chars() {
                    if enable_linewrap && cursor.line + 2 < self.text.len_lines() {
                        cursor.line += 1;
                        cursor.column = 0;
                    }
                } else {
                    cursor.column += 1;
                }
            }
        }
    }
}

#[derive(Copy, Clone, Debug, Default)]
struct Cursor {
    pub column: usize,
    pub line: usize,
}

#[derive(Copy, Clone, Debug)]
struct InsertState {
    append: bool,
}

#[derive(Copy, Clone, Debug, Default)]
enum Mode {
    #[default]
    Normal,
    Command,
    Visual,
    Insert(InsertState),
}

impl Mode {
    pub fn cursor_style(&self) -> cursor::SetCursorStyle {
        use cursor::SetCursorStyle as Style;
        match self {
            Mode::Normal => Style::SteadyBlock,
            Mode::Visual => Style::BlinkingBlock,
            Mode::Insert(_) => Style::BlinkingBar,
            Mode::Command => Style::SteadyUnderScore,
        }
    }
}

#[derive(Copy, Clone, Debug)]
enum Direction {
    Left,
    Down,
    Up,
    Right,
}

struct State {
    pub buffer: Buffer,
    pub cursor: Cursor,
    pub scroll: Cursor,
    pub size: (usize, usize),
    pub mode: Mode,
    pub quit: bool,
}

impl State {
    pub fn from_str(text: &str) -> Result<Self> {
        let (cols, rows) = terminal::size()?;

        Ok(Self {
            buffer: Buffer::from_str(text),
            cursor: Cursor::default(),
            scroll: Cursor::default(),
            size: (cols as usize, rows as usize),
            mode: Mode::default(),
            quit: false,
        })
    }

    pub fn draw(&self, out: &mut impl Write) -> Result<()> {
        out.execute(terminal::BeginSynchronizedUpdate)?;
        out.execute(terminal::Clear(terminal::ClearType::All))?;
        let lr_width = self.buffer.draw(self.scroll, out)?;
        let cursor = self.buffer.clamped_cursor(self.cursor);
        let col = cursor.column.saturating_sub(self.scroll.column) as u16;
        let row = cursor.line.saturating_sub(self.scroll.line) as u16;
        let col = col + lr_width as u16;
        out.execute(cursor::MoveTo(col, row))?;
        out.execute(self.mode.cursor_style())?;
        out.execute(terminal::EndSynchronizedUpdate)?;
        Ok(())
    }

    pub fn on_event(&mut self, event: Event) {
        match self.mode {
            Mode::Normal => self.on_normal_event(event),
            Mode::Command => self.on_command_event(event),
            Mode::Visual => self.on_visual_event(event),
            Mode::Insert(state) => self.on_insert_event(event, state),
        }
    }

    fn on_normal_event(&mut self, event: Event) {
        match event {
            Event::Key(KeyEvent { code, .. }) => match code {
                KeyCode::Char('i') => {
                    let state = InsertState { append: false }; 
                    self.mode = Mode::Insert(state);
                }
                KeyCode::Char('a') => {
                    let state = InsertState { append: true, };
                    self.move_cursor(Direction::Right);
                    self.mode = Mode::Insert(state);
                }
                KeyCode::Char(':') => {
                    self.mode = Mode::Command;
                }
                KeyCode::Char('v') => {
                    self.mode = Mode::Visual;
                }
                KeyCode::Esc => {
                    self.quit = true;
                }
                code => self.on_any_key(code),
            },
            event => self.on_any_event(event),
        }
    }

    fn on_command_event(&mut self, event: Event) {
        match event {
            Event::Key(KeyEvent { code, .. }) => match code {
                KeyCode::Esc => {
                    self.mode = Mode::Normal;
                }
                code => self.on_any_key(code),
            },
            event => self.on_any_event(event),
        }
    }

    fn on_visual_event(&mut self, event: Event) {
        match event {
            Event::Key(KeyEvent { code, .. }) => match code {
                KeyCode::Esc => {
                    self.mode = Mode::Normal;
                }
                code => self.on_any_key(code),
            },
            event => self.on_any_event(event),
        }
    }

    fn on_insert_event(&mut self, event: Event, state: InsertState) {
        match event {
            Event::Key(KeyEvent { code, .. }) => match code {
                KeyCode::Char(c) => {
                    let index = self.buffer.cursor_to_char(self.cursor);
                    self.buffer.text.insert_char(index, c);
                    self.move_cursor(Direction::Right)
                }
                KeyCode::Backspace => {
                    self.move_cursor(Direction::Left);
                    let index = self.buffer.cursor_to_char(self.cursor);
                    self.buffer.text.remove(index..=index);
                }
                KeyCode::Delete => {
                    let index = self.buffer.cursor_to_char(self.cursor);
                    self.buffer.text.remove(index..=index);
                }
                KeyCode::Enter => {
                    let index = self.buffer.cursor_to_char(self.cursor);
                    self.buffer.text.insert_char(index, '\n');
                    self.cursor.line += 1;
                    self.cursor.column = 0;
                }
                KeyCode::Esc => {
                    if state.append {
                        self.move_cursor(Direction::Left);
                    }
                    self.mode = Mode::Normal;
                }
                code => self.on_any_key(code),
            },
            event => self.on_any_event(event),
        }
    }

    fn on_any_event(&mut self, event: Event) {
        match event {
            Event::Resize(cols, rows) => {
                self.size = (cols as usize, rows as usize);
            }
            Event::Key(KeyEvent { code, .. }) => self.on_any_key(code),
            _ => {}
        }
    }

    fn on_any_key(&mut self, code: KeyCode) {
        match code {
            KeyCode::Esc => self.mode = Mode::Normal,
            KeyCode::Char('h') | KeyCode::Left => self.move_cursor(Direction::Left),
            KeyCode::Char('j') | KeyCode::Down => self.move_cursor(Direction::Down),
            KeyCode::Char('k') | KeyCode::Up => self.move_cursor(Direction::Up),
            KeyCode::Char('l') | KeyCode::Right => self.move_cursor(Direction::Right),
            _ => {}
        }
    }

    fn move_cursor(&mut self, direction: Direction) {
        self.buffer.move_cursor(&mut self.cursor, direction, true);

        if self.cursor.column < self.scroll.column + 3 {
            self.scroll.column = self.cursor.column.saturating_sub(3);
        } else if self.cursor.column + 6 >= self.scroll.column + self.size.0 {
            self.scroll.column = self.cursor.column.saturating_sub(self.size.0 - 6);
        }

        if self.cursor.line < self.scroll.line {
            self.scroll.line = self.cursor.line;
        } else if self.cursor.line + 3 >= self.scroll.line + self.size.1 {
            self.scroll.line = self.cursor.line + 3 - self.size.1;
        }
    }
}

fn screen_main(stdout: &mut Stdout, mut state: State) -> Result<()> {
    while !state.quit {
        state.draw(stdout)?;
        let event = read()?;
        state.on_event(event);
    }

    Ok(())
}

fn main() -> Result<()> {
    let argv = args().collect::<Vec<String>>();
    let mut buf = Vec::new();

    match argv.get(1).map(|s| s.as_str()) {
        Some("-") | None => {
            let stdin = 0; // get stdin as a file descriptor
            if unsafe { libc::isatty(stdin) } == 0 {
                unsafe { File::from_raw_fd(stdin) }
            } else { File::open("/dev/null").unwrap() }
        },
        Some(path) => {
            std::fs::File::open(path).unwrap_or_else(|_| {
                eprintln!(
                    "{}: {}: No such file or directory.",
                    argv[0],
                    argv[1]
                );
                exit(EX_UNAVAILABLE);
            })
        },
    }.read_to_end(&mut buf).unwrap();

    let text = String::from_utf8(buf).unwrap_or_else(|_| {
        eprintln!(
            "{}: {}: File contents are not valid UTF-8.", argv[0], argv[1]
        );
        exit(EX_DATAERR);
    });

    let state = State::from_str(&text)?;
    let mut stdout = stdout();
    terminal::enable_raw_mode()?;
    stdout.execute(terminal::EnterAlternateScreen)?;
    let result = screen_main(&mut stdout, state);
    stdout.execute(terminal::LeaveAlternateScreen)?;
    terminal::disable_raw_mode()?;
    result
}
